package pencilbox.slalom;

import java.util.Iterator;
import java.util.LinkedList;

import javax.swing.event.UndoableEditEvent;
import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;

import pencilbox.common.core.Address;
import pencilbox.common.core.BoardBase;
import pencilbox.common.core.Direction;
import pencilbox.common.core.SideAddress;
import pencilbox.util.ArrayUtil;

/**
 * uX[vՖʃNX
 */
public class Board extends BoardBase {

	private static final int HORIZ = Direction.HORIZ;
	private static final int VERT = Direction.VERT;

	//Ֆ
	//}XFȂ
	//}XF[U͂BʂȂƂ
	//stateɓl
	static final int UNKNOWN = 0;
	static final int LINE = 1;
	static final int NOLINE = -1; // GUIł͕sgpH
	static final int OUTER = -9;

	//number ɓl
	//0ȏ͍}X
	/** n_ */
	static final int START = -1;
	/**  */
	static final int BLANK = -2;

	//F

	/** }X̏ */
	private int[][] number;
	/** ӂ̏ */
	private int[][][] state;
	/** ̃Xg */
	private LinkedList<Gate> gates;
	/** c̊ */
	private Gate[][] vGates;
	/** ̊ */
	private Gate[][] hGates;

	/** ׂẴC */
	private LinkedList<Link> linkList;
	/** c̃C */
	private Link[][][] link;
	private Link initializingLink;

	protected void setup() {
		super.setup();
		number = new int[rows()][cols()];
		ArrayUtil.initArrayInt2(number, BLANK);
		state = new int[2][][];
		state[VERT] = new int[rows()][cols() - 1];
		state[HORIZ] = new int[rows() - 1][cols()];
		linkList = new LinkedList<Link>();
		link = new Link[2][][];
		link[VERT] = new Link[rows()][cols() - 1];
		link[HORIZ] = new Link[rows() - 1][cols()];
		gates = new LinkedList<Gate>();
		vGates = new Gate[rows()][cols()];
		hGates = new Gate[rows()][cols()];
	}

	/**
	 * ̍W}XǂB
	 * @param r sW
	 * @param c W
	 * @return }XȂ true ԂB
	 */
	public boolean isBlack(int r, int c) {
		return isOn(r,c) && number[r][c] >= 0;
	}
	/**
	 * w肵}Xɍ}XA󔒃}X̏Ԃݒ肷
	 * @param r sW
	 * @param c W
	 * @param n ݒ肷
	 */
	public void setNumber(int r, int c, int n) {
		int old = number[r][c];
		if (n == old) {
			return;
		}

		if (old < 0 && n >= 0) {
			//}X}Xɂꍇ
			//傪Ȃ܂
			eraseGate(r, c);
		} else if (old >= 0 && n < 0) {
			//}X𔒃}Xɂꍇ
			//̍}X̂тĂ폜܂
			if (r > 0) {
				Gate vGate = vGates[r - 1][c];
				if (vGate != null) {
					removeVGate(vGate);
				}
			}
			if (r < rows() - 1) {
				Gate vGate = vGates[r + 1][c];
				if (vGate != null) {
					removeVGate(vGate);
				}
			}
			if (c > 0) {
				Gate hGate = hGates[r][c - 1];
				if (hGate != null) {
					removeHGate(hGate);
				}
			}
			if (c < cols() - 1) {
				Gate hGate = hGates[r][c + 1];
				if (hGate != null) {
					removeHGate(hGate);
				}
			}
		}

		number[r][c] = n;
	}

	public void setNumber(Address pos, int n) {
		setNumber(pos.r(), pos.c(), n);
	}
	/**
	 * w肵}X̏Ԃ擾
	 * @param r sW
	 * @param c W
	 */
	public int getNumber(int r, int c) {
		return number[r][c];
	}

	public int getNumber(Address pos) {
		return getNumber(pos.r(), pos.c());
	}

	public void setStart(int r, int c) {
		//݂̃}XX^[gǂ`FbN
		if (number[r][c] == START) {
			return;
		}
		//
		for (int row = 0; row < number.length; ++row) {
			int[] array = number[row];
			for (int col = 0; col < array.length; ++col) {
				if (array[col] == START) {
					array[col] = BLANK;
				}
			}
		}
		setNumber(r, c, START);
	}

	/**
	 * W̗2}Xꂩ}XȂ}Xǂ
	 * @param d
	 * @param r
	 * @param c
	 * @return W̗2}Xꂩ}XȂ}Xł true
	 */
	private boolean hasBlack(int d, int r, int c) {
		if (d == VERT)
			return isBlack(r, c) || isBlack(r, c+1);
		else if (d == HORIZ)
			return isBlack(r, c) || isBlack(r+1, c);
		return false;
	}
	/**
	 * 󔒃}Xɂ
	 * @param r sW
	 * @param c W
	 */
	public void eraseNumber(int r, int c) {
		setNumber(r, c, BLANK);
	}
	/**
	 * ̓͂󂯕t
	 *
	 * @param r sW
	 * @param c W
	 * @param n ͂ꂽ
	 */
	public void enterNumber(int r, int c, int n) {
		eraseLinesAround(r,c);
		setNumber(r, c, n);
	}
	/**
	 * @return Returns the state.
	 */
	int[][][] getState() {
		return state;
	}
	/**
	 * ӏԂ̎擾
	 * @param d
	 * @param r
	 * @param c
	 * @return ӂ̏ԂԂ
	 */
	public int getState(int d, int r, int c) {
		if (isSideOn(d,r,c))
			return state[d][r][c];
		else
			return OUTER;
	}
	/**
	 * ӏԂ̎擾B}XƌōWw肷B
	 * @param pos
	 * @param d
	 * @return
	 */
	public int getStateJ(Address pos, int d) {
		switch (d) {
			case Direction.UP :
				return getState(HORIZ, pos.r()-1, pos.c());
			case Direction.LT :
				return getState(VERT, pos.r(), pos.c()-1);
			case Direction.DN :
				return getState(HORIZ, pos.r(), pos.c());
			case Direction.RT :
				return getState(VERT, pos.r(), pos.c());
			default:
				return OUTER;
		}
	}
	/**
	 * ӏԂ̐ݒ
	 * @param d
	 * @param r
	 * @param c
	 * @param st
	 */
	public void setState(int d, int r, int c, int st) {
		if (isSideOn(d, r, c))
			state[d][r][c] = st;
	}

	/**
	 * ӏԂ̐ݒB}XƌōWw肷B
	 * @param pos
	 * @param d
	 * @param st
	 */
	public void setStateJ(Address pos, int d, int st) {
		switch (d) {
			case Direction.UP :
				 setState(HORIZ, pos.r()-1, pos.c(), st);
				 break;
			case Direction.LT :
				 setState(VERT, pos.r(), pos.c()-1, st);
				 break;
			case Direction.DN :
				 setState(HORIZ, pos.r(), pos.c(), st);
				 break;
			case Direction.RT :
				 setState(VERT, pos.r(), pos.c(), st);
				 break;
			default:
		}
	}

	public boolean isLine(int d, int r, int c) {
		if (!isSideOn(d,r,c)) return false;
		return state[d][r][c] == LINE;
	}

	public Link getLink(int d, int r, int c) {
		if (isSideOn(d,r,c)) return link[d][r][c];
		return null;
	}
	public Link getLink(SideAddress pos) {
		return link[pos.d()][pos.r()][pos.c()];
	}
	/**
	 * ̃}X܂ Link Ԃ
	 */
	public Link getLink(int r, int c) {
		Link link;
		link = getLink(VERT, r, c - 1);
		if (link != null) return link;
		link = getLink(VERT, r, c);
		if (link != null) return link;
		link = getLink(HORIZ, r - 1, c);
		if (link != null) return link;
		link = getLink(HORIZ, r, c);
		if (link != null) return link;
		return null;
	}
	public void setLink(int d, int r, int c, Link l) {
		link[d][r][c] =  l;
	}
	public void setLink(SideAddress pos, Link l) {
		link[pos.d()][pos.r()][pos.c()] =  l;
	}
	/**
	 * Ֆʂ Link 邩
	 * @return@Link Ȃ true
	 */
	public boolean hasMultipleLinks() {
		return linkList.size() > 1;
	}

	/**
	 * }X㉺E4ɈĂ
	 * }X}X␔}XɕύXꂽꍇɐ邽߂Ɏgp
	 * @param r }X̍sW
	 * @param c }X̗W
	 */
	void eraseLinesAround(int r, int c) {
		if (getState(VERT, r, c-1) == LINE)
			changeStateA(VERT, r, c-1, UNKNOWN);
		if (getState(VERT, r, c) == LINE)
			changeStateA(VERT, r, c, UNKNOWN);
		if (getState(HORIZ, r, c) == LINE)
			changeStateA(HORIZ, r, c, UNKNOWN);
		if (getState(HORIZ, r-1, c) == LINE)
			changeStateA(HORIZ, r-1, c, UNKNOWN);
	}
	/**
	 * ՖʏԂύXCAhDXi[ɕύXʒmD
	 * @param r
	 * @param c
	 * @param st
	 */
	public void changeStateA(int r, int c, int st) {
		fireUndoableEditUpdate(
			new UndoableEditEvent(this, new PaintStep(r, c, number[r][c], st)));
		setNumber(r, c, st);
	}
	/**
	 * ӂ̏Ԃw肵ԂɕύX
	 * @param d c
	 * @param r sW
	 * @param c W
	 * @param st ύX̏
	 */
	public void changeState(int d, int r, int c, int st) {
		int previousState = getState(d,r,c);
		setState(d,r,c,st);
		if (previousState == LINE) {
			cutLink(d,r,c);
		}
		if (st == LINE) {
			connectLink(d,r,c);
		}
	}
	/**
	 * ӂ̏Ԃw肵ԂɕύX
	 * AhDXi[ɕύXʒm
	 * @param d c
	 * @param r sW
	 * @param c W
	 * @param st ύX̏
	 */
	public void changeStateA(int d, int r, int c, int st) {
		fireUndoableEditUpdate(
			new UndoableEditEvent(this, new LineStep(d, r, c, state[d][r][c], st)));
		changeState(d, r, c, st);
	}
	/**
	 * ӂ̏Ԃ st Ő؂ւ
	 * @param d c
	 * @param r sW
	 * @param c W
	 * @param st ؂ւ
	 */
	public void toggleState(int d, int r, int c, int st) {
		if (getState(d, r, c) == st)
			changeStateA(d, r, c, UNKNOWN);
		else
			changeStateA(d, r, c, st);
	}

	/**
	 * n_}XƏI_}X񂾐̏Ԃw̏ԂɕύX
	 * n_}XƏI_}X͓s܂͓ɂȂ΂ȂȂ
	 * @param pos0 n_}X̍W
	 * @param pos1 I_}X̍W
	 * @param st ύX̏
	 */
	public void determineInlineState(Address pos0, Address pos1, int st) {
		int ra = pos0.r()<pos1.r() ? pos0.r() : pos1.r();
		int rb = pos0.r()<pos1.r() ? pos1.r() : pos0.r();
		int ca = pos0.c()<pos1.c() ? pos0.c() : pos1.c();
		int cb = pos0.c()<pos1.c() ? pos1.c() : pos0.c();
		if (ra == rb)
			for (int c = ca; c < cb; c++) {
				if (getState(VERT, ra, c) != st)
					if (!hasBlack(VERT, ra, c) || st == UNKNOWN)
						changeStateA(VERT, ra, c, st);
			}
		if (ca == cb)
			for (int r = ra; r < rb; r++) {
				if (getState(HORIZ, r, ca) != st)
					if (!hasBlack(HORIZ, r, ca) || st == UNKNOWN)
						changeStateA(HORIZ, r, ca, st);
			}
	}

	public void clearBoard() {
		super.clearBoard();
		ArrayUtil.initArrayInt3(state, UNKNOWN);
//		for (int r=0; r<rows(); r++) {
//			for (int c=0; c<cols(); c++) {
//				if (!isNumber(r,c)) {
//					number[r][c] = BLANK;
//				}
//			}
//		}
		initBoard();
	}

	public void trimAnswer() {
//		for (int r=0; r<rows(); r++) {
//			for (int c=0; c<cols(); c++) {
//				if (getNumber(r, c) == WHITE)
//					setNumber(r, c, BLANK);
//				}
//		}
	}

	public void initBoard() {
		initLinks();
	}

	void initLinks() {
		Link.resetID();
		linkList.clear();
		ArrayUtil.initArrayObject2(link[0],null);
		ArrayUtil.initArrayObject2(link[1],null);
		for (int r=0; r<rows(); r++) {
			for (int c=0; c<cols(); c++) {
				initLink(r, c);
			}
		}
	}

	/**
	 * ̃}X̏㉺E̗אڂS}Xɍ}X邩ǂ𒲂ׂ
	 * @param r
	 * @param c
	 * @return ㉺Eɍ}XЂƂł true
	 */
	boolean isBlock(int r, int c) {
		if (isBlack(r-1, c) || isBlack(r+1, c) || isBlack(r, c-1) || isBlack(r, c+1))
			return true;
		return false;
	}

	/**
	 * }X܂ Link ̏
	 * link[][][] ͏Ă̂Ƃ
	 * @param r Link̋N_}X̍sW
	 * @param c Link̋N_}X̗W
	 */
	void initLink (int r, int c) {
		initializingLink = new Link();
		initLink1(VERT , r  , c-1);
		initLink1(VERT , r  , c  );
		initLink1(HORIZ, r-1, c  );
		initLink1(HORIZ, r  , c  );
		if (!initializingLink.isEmpty())
			linkList.add(initializingLink);
//		printLink(d,r,c);
	}
	private void initLink1(int d, int r, int c) {
		if (!isSideOn(d,r,c)) return;
		if (!isLine(d,r,c)) return;
		if (getLink(d,r,c) != null) return;
		initializingLink.add(d,r,c);
		setLink(d, r, c, initializingLink);
		if (d==VERT) {
			initLink1(VERT , r  , c-1);
			initLink1(VERT , r  , c+1);
			initLink1(HORIZ, r-1, c  );
			initLink1(HORIZ, r-1, c+1);
			initLink1(HORIZ, r  , c  );
			initLink1(HORIZ, r  , c+1);
		}
		if (d==HORIZ) {
			initLink1(HORIZ, r-1, c  );
			initLink1(HORIZ, r+1, c  );
			initLink1(VERT , r  , c-1);
			initLink1(VERT , r+1, c-1);
			initLink1(VERT , r  , c  );
			initLink1(VERT , r+1, c  );
		}
	}
	/**
	 * Link 
	 */
	void connectLink(int d, int r, int c) {
		Link newLink = null;
		Link link1 = null;
		Link link2 = null;
		if (d==VERT) {
			link1 = getLink(r,c);
			link2 = getLink(r,c+1);
		} else if (d==HORIZ) {
			link1 = getLink(r,c);
			link2 = getLink(r+1,c);
		}
		if (link1==null && link2 == null) {
			newLink = new Link();
			linkList.add(newLink);
		} else if (link1==null && link2!=null) {
			newLink = link2;
		} else if (link1!=null && link2==null) {
			newLink = link1;
		} else if (link1==link2) {
			newLink = link1;
		} else {
			if (link1.size() >= link2.size()) {
				newLink = link1;
				newLink.addAll(link2);
				for(SideAddress joint : link2) {
					setLink(joint, newLink);
				}
				linkList.remove(link2);
			}
			else {
				newLink = link2;
				newLink.addAll(link1);
				for(SideAddress joint : link1) {
					setLink(joint, newLink);
				}
				linkList.remove(link1);
			}
		}
		newLink.add(d,r,c);
		setLink(d,r,c, newLink);
//		printLink(d,r,c);
	}


	/**
	 * Link ؒf
	 */
	void cutLink(int d, int r, int c) {
		Link oldLink = getLink(d,r,c);
		Link longerLink = null;
		for (SideAddress joint : oldLink) {
			setLink(joint, null);
		}
		linkList.remove(oldLink);
		if (d==VERT) {
			initLink(r  , c  );
			longerLink = initializingLink;
			initLink(r  , c+1);
			if (initializingLink.size() > longerLink.size()) longerLink = initializingLink;
		}
		else if (d==HORIZ) {
			initLink(r  , c  );
			longerLink = initializingLink;
			initLink(r+1, c  );
			if (initializingLink.size() > longerLink.size()) longerLink = initializingLink;
		}
		longerLink.setID(oldLink.getID());
	}

	/**
	 * }X̏㉺E4̂CݐĂ鐔Ԃ
	 * @param r }X̍sW
	 * @param c }X̗W
	 * @return }X̏㉺EɈĂ̐
	 */
	public int countLine(int r, int c) {
		int no = 0;
		if (r < rows() - 1 && isLine(HORIZ, r, c))
			no++;
		if (c < cols() - 1 && isLine(VERT, r, c))
			no++;
		if (r > 0 && isLine(HORIZ, r - 1, c))
			no++;
		if (c > 0 && isLine(VERT, r, c - 1))
			no++;
		return no;
	}

	public boolean existLine(int r, int c) {
		if (r < rows() - 1 && isLine(HORIZ, r, c))
			return true;
		if (c < cols() - 1 && isLine(VERT, r, c))
			return true;
		if (r > 0 && isLine(HORIZ, r - 1, c))
			return true;
		if (c > 0 && isLine(VERT, r, c - 1))
			return true;
		return false;
	}

	/**
	 * ̗[̂ꂩ̈ʒuŐ򂵂Ă邩ǂ𒲂ׂ
	 * @param d ӈʒu͏c
	 * @param r ӂ̍sW
	 * @param c ӂ̗W
	 * @return [̂ꂩŐ򂵂Ă true 򂵂ĂȂ false Ԃ
	 */
	boolean isBranchedLink(int d, int r, int c) {
		if (countLine(r,c) > 2)
			return true;
		if (d==VERT) {
			if (countLine(r,c+1) > 2)
				return true;
		} else if (d==HORIZ) {
			if (countLine(r+1,c) > 2)
				return true;
		}
		return false;
	}

	/**
	 * }XƏdȂĂ邩𒲂ׂ
	 * @param d ӈʒu͏c
	 * @param r ӂ̍sW
	 * @param c ӂ̗W
	 * @return }XƏdȂĂtrueԂ
	 */
	boolean isBuriedLink(int d, int r, int c) {
		if (d==VERT) {
			if (isBlack(r, c) || isBlack(r, c+1))
				return true;
		} else if (d==HORIZ) {
			if (isBlack(r, c) || isBlack(r+1, c))
				return true;
		}
		return false;
	}

	/**
	 * Ñ`FbNs
	 */
	private int checkLinks() {
		int result = 0;
		for (int r = 0; r < rows(); r++) {
			for (int c = 0; c < cols(); c++) {
				int l = countLine(r,c);
				if (l > 2) {
					result |= 1;
				} else if ( l == 1 ) {
					result |= 2;
				}
				if (isBlack(r,c) && (l > 0))
					result |= 8;
//				if (!isBlack(r,c) && (l == 0))
//					result |= 8;
			}
		}
		if (hasMultipleLinks())
			result |= 4;
		return result;
	}

	public int checkAnswerCode() {
		int result = 0;
		result |= checkProblem();
		if (result == 0) {
			result |= checkLinks();
		}
		if (result == 0) {
			result |= checkGates();
		}
		return result;
	}

	private int checkProblem() {
		//ƂĐǂ̃`FbNs
		//n_ȂƂ`FbN
		//̐傫ȂƂ`FbN
		//̐LтĂȂ̂}X݂ȂƂ`FbN
		int result = 0;
		int startCount = 0;
		int gateCount = gates.size();

		//̐傫ȂƂ`FbN
		//̐LтĂȂ̂}X݂ȂƂ`FbN
		for (int row = 0; row < number.length; ++row) {
			int[] array = number[row];
			for (int col = 0; col < array.length; ++col) {
				int num = array[col];
				if (num == START) {
					++startCount;
				} else if (num > 0) {
					if (num > gateCount) {
						result |= (1 << 18);
					}
					//
					boolean notfound = true;
					if (notfound && row > 0) {
						Gate vGate = vGates[row - 1][col];
						if (vGate != null) {
							notfound = false;
						}
					}
					if (notfound && row < rows() - 1) {
						Gate vGate = vGates[row + 1][col];
						if (vGate != null) {
							notfound = false;
						}
					}
					if (notfound && col > 0) {
						Gate hGate = hGates[row][col - 1];
						if (hGate != null) {
							notfound = false;
						}
					}
					if (notfound && col < cols() - 1) {
						Gate hGate = hGates[row][col + 1];
						if (hGate != null) {
							notfound = false;
						}
					}
					if (notfound) {
						result |= (1 << 19);
					}
				}
			}
		}

		//n_ȂƂ`FbN
		if (startCount == 0) {
			result |= (1 << 16);
		} else if (startCount > 1) {
			result |= (1 << 17);
		}

		return result;
	}

	/**
	 * @return
	 */
	private int checkGates() {
		//őOƂăNƃNɂȂĂ̂Ƃ
		//n_݂ꍇAn_ʉ߂Ă邩ǂ`FbN
		int result = 0;

		int startRow = 0;
		int startCol = 0;
		for (int row = 0; row < number.length; ++row) {
			int[] array = number[row];
			for (int col = 0; col < array.length; ++col) {
				if (array[col] == START) {
					startRow = row;
					startCol = col;
					if (!existLine(row, col)) {
						result |= (1 << 8);
					}
				}
			}
		}
		//̏}X݂邩`FbN
		//1łȂƃ_
		Iterator<Gate> iterator = gates.iterator();

		while (iterator.hasNext()) {
			Gate gate = iterator.next();

			int count = 0;
			if (gate.direct == VERT) {
				for (int y = gate.startRow; y <= gate.endRow; ++y) {
					if (existLine(y, gate.startCol)) {
						++count;
					}
				}
			} else {
				for (int x = gate.startCol; x <= gate.endCol; ++x) {
					if (existLine(gate.startRow, x)) {
						++count;
					}
				}
			}

			if (count == 0) {
				result |= (1 << 9);
			} else if (count > 1) {
				result |= (1 << 10);
			}
		}

		if (result == 0) {
			//ɒʉ߂ԍݒ肵܂
			int row = startRow;
			int col = startCol;
			int count = 0;
			int direction = 0;

			while (true) {
				if (direction != 3 && row > 0 && isLine(HORIZ, row - 1, col)) {
					//
					row -= 1;
					direction = 1;
				} else if (direction != 1 && row < rows() - 1 && isLine(HORIZ, row, col)) {
					//
					row += 1;
					direction = 3;
				} else if (direction != 4 && col > 0 && isLine(VERT, row, col - 1)) {
					//
					col -= 1;
					direction = 2;
				} else if (direction != 2 && col < cols() - 1 && isLine(VERT, row, col)) {
					//E
					col += 1;
					direction = 4;
				} else {
					throw new RuntimeException();
				}

				if (vGates[row][col] != null) {
					vGates[row][col].number = ++count;
				}
				if (hGates[row][col] != null) {
					hGates[row][col].number = ++count;
				}

				if (row == startRow && col == startCol) {
					break;
				}
			}

			//Ƌt̃`FbNsBÂꂩȂOK
			boolean normal = true;
			boolean reverse = true;
			for (row = 0; row < number.length; ++row) {
				int[] array = number[row];
				for (col = 0; col < array.length; ++col) {
					int num = array[col];
					if (num > 0) {
						boolean workNormal = false;
						boolean workReverse = false;
						if (row > 0) {
							Gate vGate = vGates[row - 1][col];
							if (vGate != null) {
								if (vGate.number == num) {
									workNormal = true;
								}
								if ((count + 1 - vGate.number) == num) {
									workReverse = true;
								}
							}
						}
						if (row < rows() - 1) {
							Gate vGate = vGates[row + 1][col];
							if (vGate != null) {
								if (vGate.number == num) {
									workNormal = true;
								}
								if ((count + 1 - vGate.number) == num) {
									workReverse = true;
								}
							}
						}
						if (col > 0) {
							Gate hGate = hGates[row][col - 1];
							if (hGate != null) {
								if (hGate.number == num) {
									workNormal = true;
								}
								if ((count + 1 - hGate.number) == num) {
									workReverse = true;
								}
							}
						}
						if (col < cols() - 1) {
							Gate hGate = hGates[row][col + 1];
							if (hGate != null) {
								if (hGate.number == num) {
									workNormal = true;
								}
								if ((count + 1 - hGate.number) == num) {
									workReverse = true;
								}
							}
						}

						if (!workNormal) {
							normal = false;
						}
						if (!workReverse) {
							reverse = false;
						}
					}
				}
			}

			if (!normal && !reverse) {
				result |= (1 << 11);
			}
		}

		return result;
	}

	static final String[] ANSWER_STRINGS = {
		"܂͌Ă",
		"ĂȂ",
		"̐",
		"ƍ}XdȂĂ}X",
		"",
		"",
		"",
		"",
		"n_ʂĂȂ",
		"ʂĂȂ傪",
		"2}XȏʂĂ傪",
		"ʂɊʉ߂Ă܂",
		"",
		"",
		"",
		"",
		"n_݂܂",
		"n_݂܂",
		"̐傫Ȑ̓Ă}X܂",
		"傪LтĂȂ̓Ă}X܂",
		"",
		"",
		"",
		"",
	};


	public String checkAnswerString() {
		int result = checkAnswerCode();
		if (result == 0)
			return COMPLETE_MESSAGE;
		StringBuffer message = new StringBuffer();

		for (int i = 0; i < ANSWER_STRINGS.length; ++i) {
			if ((result & (1 << i)) != 0) {
				message.append(ANSWER_STRINGS[i]);
				message.append("\n");
			}
		}

		return message.toString();
	}

	/**
	 * P̑\NX
	 * UNDO, REDO ł̕ҏW̒PʂƂȂ
	 */
	class Step extends AbstractUndoableEdit {
	}

	/**
	 * P̑\NX
	 * UNDO, REDO ł̕ҏW̒PʂƂȂ
	 */
	class LineStep extends Step {

		private int direction;
		private int row;
		private int col;
		private int before;
		private int after;
		/**
		 * RXgN^
		 * @param d c
		 * @param r ύXꂽ}X̍sW
		 * @param c ύXꂽ}X̗W
		 * @param b ύXȌ
		 * @param a ύX̏
		 */
		public LineStep(int d, int r, int c, int b, int a) {
			super();
				direction = d;
			row = r;
			col = c;
			before = b;
			after = a;
		}
		public void undo() throws CannotUndoException {
			super.undo();
			changeState(direction, row, col, before);
		}
		public void redo() throws CannotRedoException {
			super.redo();
			changeState(direction, row, col, after);
		}
	}
	/**
	 * P̑\NX
	 * UNDO, REDO ł̕ҏW̒PʂƂȂ
	 */
	class PaintStep extends Step {

		private int row;
		private int col;
		private int before;
		private int after;
		/**
		 * RXgN^
		 * @param r ύXꂽ}X̍sW
		 * @param c ύXꂽ}X̗W
		 * @param b ύXȌ
		 * @param a ύX̏
		 */
		public PaintStep(int r, int c, int b, int a) {
			super();
			row = r;
			col = c;
			before = b;
			after = a;
		}
		public void undo() throws CannotUndoException {
			super.undo();
			setNumber(row, col, before);
		}
		public void redo() throws CannotRedoException {
			super.redo();
			setNumber(row, col, after);
		}
	}

	/**
	 * \܂
	 */
	static class Gate {
		/**  */
		int direct;
		/** n_ */
		int startRow;
		int startCol;
		/** I_ */
		int endRow;
		int endCol;
		/** ԍ */
		int number;

		public Gate(int direct, int base, int start, int end) {
			this.direct = direct;
			if (direct == HORIZ) {
				startRow = base;
				startCol = start;
				endRow = base;
				endCol = end;
			} else {
				startCol = base;
				startRow = start;
				endCol = base;
				endRow = end;
			}
		}
	}

	public void addHGate(int r, int c) {
		//̈ʒu}XłȂA傪ȂƂ`FbN
		if (number[r][c] >= 0 || hGates[r][c] != null) {
			return;
		}
		//n_ƏI_`FbN܂
		int start = c;
		while (start > 0) {
			if (number[r][start - 1] >= 0) {
				break;
			}
			--start;
		}
		int end = c;
		while (end < cols() - 1) {
			if (number[r][end + 1] >= 0) {
				break;
			}
			++end;
		}
		Gate gate = new Gate(HORIZ, r, start, end);
		gates.add(gate);
		for (int col = start; col <= end; ++col) {
			hGates[r][col] = gate;
		}
	}

	public void addVGate(int r, int c) {
		//̈ʒu}XłȂA傪ȂƂ`FbN
		if (number[r][c] >= 0 || vGates[r][c] != null) {
			return;
		}
		//n_ƏI_`FbN܂
		int start = r;
		while (start > 0) {
			if (number[start - 1][c] >= 0) {
				break;
			}
			--start;
		}
		int end = r;
		while (end < rows() - 1) {
			if (number[end + 1][c] >= 0) {
				break;
			}
			++end;
		}
		Gate gate = new Gate(VERT, c, start, end);
		gates.add(gate);
		for (int row = start; row <= end; ++row) {
			vGates[row][c] = gate;
		}
	}

	public int getGateCount() {
		return gates.size();
	}

	public Gate getHGate(int r, int c) {
		return hGates[r][c];
	}

	public Gate getVGate(int r, int c) {
		return vGates[r][c];
	}

	private void removeHGate(Gate gate) {
		gates.remove(gate);
		int r = gate.startRow;
		int start = gate.startCol;
		int end = gate.endCol;
		for (int col = start; col <= end; ++col) {
			hGates[r][col] = null;
		}
	}

	private void removeVGate(Gate gate) {
		gates.remove(gate);
		int c = gate.startCol;
		int start = gate.startRow;
		int end = gate.endRow;
		for (int row = start; row <= end; ++row) {
			vGates[row][c] = null;
		}
	}

	public void eraseGate(int r, int c) {
		Gate hGate = hGates[r][c];
		if (hGate != null) {
			removeHGate(hGate);
		}
		Gate vGate = vGates[r][c];
		if (vGate != null) {
			removeVGate(vGate);
		}
	}

	public boolean exist(int r, int c) {
		if (number[r][c] >= START) return true;
		if (hGates[r][c] != null) return true;
		if (vGates[r][c] != null) return true;
		return false;
	}

	public void clear(int r, int c) {
		eraseNumber(r, c);
		eraseGate(r, c);
	}

	public Iterator<Gate> getGateIterator() {
		return gates.iterator();
	}
}
