package pencilbox.slalom;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Stroke;
import java.awt.geom.Line2D;

import pencilbox.common.core.BoardBase;
import pencilbox.common.gui.PanelBase;
import pencilbox.util.Colors;


/**
 * uX[vplNX
 */
public class Panel extends PanelBase {

	private Board board;

	private boolean indicateErrorMode = false;
	private boolean separateLinkColorMode = false;

	private Color lineColor = Color.BLUE;
	private Color paintColor = Color.BLACK;
	private Color startColor = Color.BLACK;
	private Color gateColor = Color.BLACK;

	/**
	 *
	 */
	public Panel() {
		super();
		setNumberColor(Color.WHITE);
		setGridColor(Color.BLACK);
		setMarkStyle(3);
	}

	protected void setBoard(BoardBase aBoard) {
		board = (Board) aBoard;
	}

	/**
	 * @return Returns the separateLinkColorMode.
	 */
	public boolean isSeparateLinkColorMode() {
		return separateLinkColorMode;
	}

	/**
	 * @param separateLinkColorMode The separateLinkColorMode to set.
	 */
	public void setSeparateLinkColorMode(boolean separateLinkColorMode) {
		this.separateLinkColorMode = separateLinkColorMode;
	}

	/**
	 * @return Returns the indicateErrorMode.
	 */
	public boolean isIndicateErrorMode() {
		return indicateErrorMode;
	}

	/**
	 * @param indicateErrorMode The indicateErrorMode to set.
	 */
	public void setIndicateErrorMode(boolean indicateErrorMode) {
		this.indicateErrorMode = indicateErrorMode;
	}

	/**
	 * @return Returns the blackColor.
	 */
	public Color getPaintColor() {
		return paintColor;
	}

	/**
	 * @param blackColor The blackColor to set.
	 */
	public void setPaintColor(Color blackColor) {
		this.paintColor = blackColor;
	}

	/**
	 * @return Returns the lineColor.
	 */
	public Color getLineColor() {
		return lineColor;
	}

	/**
	 * @param lineColor The lineColor to set.
	 */
	public void setLineColor(Color lineColor) {
		this.lineColor = lineColor;
	}

	/**
	 * @param startColor the startColor to set
	 */
	public void setStartColor(Color startColor) {
		this.startColor = startColor;
	}

	/**
	 * @return the startColor
	 */
	public Color getStartColor() {
		return startColor;
	}
	/**
	 * @param gateColor the gateColor to set
	 */
	public void setGateColor(Color gateColor) {
		this.gateColor = gateColor;
	}

	/**
	 * @return the gateColor
	 */
	public Color getGateColor() {
		return gateColor;
	}

	public void drawBoard(Graphics2D g) {
		//Z`悵܂
		drawNumbers(g);
		//`悵܂
		drawGate(g);
		//r`悵܂
		drawGrid(g);
		//C`悵܂
		drawLinks(g);
		//Og`悵܂
		drawBoardBorder(g);
	}

	private void drawNumbers(Graphics2D g) {
		g.setFont(getNumberFont());
		for (int r = 0; r < board.rows(); r++) {
			for (int c = 0; c < board.cols(); c++) {
				int state = board.getNumber(r, c);
				if (state >= 0) {
					g.setColor(getPaintColor());
					paintCell(g, r, c);
					if (state > 0) {
						g.setColor(getNumberColor());
						placeNumber(g, r, c, state);
					}
				} else if (state == Board.START) {
					//ۂ܂
					g.setColor(getStartColor());
					placeCircle(g, r, c, getCellSize() - 3);
					int num = board.getGateCount();
					if (num > 0) {
						placeNumber(g, r, c, num);
					}
					placeCircle(g, r, c, getCellSize() - 1);
				}
//				if (state == Board.WHITE) {
//					g.setColor(getCircleColor());
//					placeMark(g, r, c);
//				}
			}
		}
	}

	private void drawGate(Graphics2D g) {
		g.setColor(getGateColor());
		Line2D.Float line = new Line2D.Float();
		float dash = ((float)getCellSize()) / 4;
		float[] dashArray = {dash, dash};
		Stroke oldStroke = g.getStroke();
		g.setStroke(new BasicStroke(3.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 0f, dashArray, dash / 2));
		for (int r = 0; r < board.rows(); r++) {
			for (int c = 0; c < board.cols(); c++) {
				if (board.getHGate(r, c) != null) {
					line.setLine(toX(c), toY(r) + getHalfCellSize(), toX(c) + getCellSize(), toY(r) + getHalfCellSize());
					g.draw(line);
				}
				if (board.getVGate(r, c) != null) {
					line.setLine(toX(c) + getHalfCellSize(), toY(r), toX(c) + getHalfCellSize(), toY(r) + getCellSize());
					g.draw(line);
				}
			}
		}
		g.setStroke(oldStroke);
	}

	private void drawLinks(Graphics2D g) {
		int state;
		for (int d = 0; d <= 1; d++) {
			for (int r = 0; r < board.rows(); r++) {
				for (int c = 0; c < board.cols(); c++) {
					state = board.getState(d, r, c);
					if (state == Board.LINE) {
						g.setColor(getLineColor());
						if (isSeparateLinkColorMode())
							g.setColor(Colors.getColor(board.getLink(d,r,c).getID()));
						if (isIndicateErrorMode()) {
							if (board.isBranchedLink(d,r,c))
								g.setColor(getErrorColor());
							else if (board.isBuriedLink(d,r,c))
								g.setColor(getErrorColor());
						}
						placeLink(g, d, r, c);
//					} else if (state == Board.NOLINE) {
//						g.setColor(getCrossColor());
//						placeSideCross(g, d, r, c);
					}
				}
			}
		}
	}
}
